package homegear;

import java.io.*;
import java.util.*;
import java.text.*;

/**
  * Representation of the homegear RPC arrays and structs. 
  * A Usage demonstration is given in {@link #main(String[])}.
  */

public class HomeGearArray {
    private boolean isNamed;		// if fields have a name it is a struct
    private int size;			// number of elements
    private int[] valPos;		// position of the element values
    private int[] namePos;		// position of the element names names of 0 if there are no names
    private HomeGearArray[] child;	// child arrays

    /* * data buffer encoded according to HomeGear binary standard. The buffer should not modified by user. */
    protected byte[] data;	
    /* * Start of the buffer. The buffer starts with the number of elements. This field should not be modified by user.*/
    protected int offs;	
    /* * Amount of bytes used for this struct. Actual size may be larger. This field should not be modified by user.*/
    protected int length;	

    /** * Integer identifier */
    public static final int TYPE_INT = 1;
    /** * Boolean identifier */
    public static final int TYPE_BOOL = 2;
    /** * String identifier */
    public static final int TYPE_STRING = 3;
    /** * Float identifier */
    public static final int TYPE_FLOAT = 4;
    /** * BASE64 identifier */
    public static final int TYPE_BASE64 = 0x11;
    /** * Array identifier */
    public static final int TYPE_ARRAY = 0x100;
    /** * Struct identifier */
    public static final int TYPE_STRUCT = 0x101;

/**
  * Constructs an array or a struct from a list of variables
  * @param names The field names. If length of names is > 0 and names != null its a struct, otherwise an array.
  * @param vals Values of the struct or array.
  * @throws IllegalArgumentException if a parameter is invalid.
  * @throws IOException if an I/O error occurs.
  */
    public HomeGearArray(String[] names, Object[] vals) throws IllegalArgumentException,IOException  {
	isNamed = ( names != null ) && (names.length > 0);
	offs = 0;
	if ( vals == null || vals.length==0 ) {  // empty arrays are allowed
	    isNamed = false;
	    size = 0;
	    data = new byte[] { 0,0,0,0 };
	    length = 4;
	    valPos = null;
	    namePos = null;
	    child = null;
	    return;
	}
	if (isNamed && names.length != vals.length ) throw new IllegalArgumentException("Unequal number of names and number of values");
	
	size = vals.length;
	valPos = new int[size];
	namePos = new int[size];
	child = new HomeGearArray[size];

	ByteArrayOutputStream buf = new ByteArrayOutputStream();
	DataOutputStream writer = new DataOutputStream(buf);
	writer.writeInt(size);			// field count
	for (int i=0; i<vals.length; i++) {
	    // name
	    if ( isNamed ) {
		namePos[i] = buf.size();
		writer.writeInt(names[i].length());
		writer.writeBytes(names[i]);
	    }	
	    else {
		namePos[i] = 0;
	    }
	    
	    // value
	    child[i] = null;
	    valPos[i] = buf.size();
	    if ( vals[i] == null ) {
		writer.writeInt(TYPE_STRING);
		writer.writeInt(0);
	    }
	    else if ((vals[i] instanceof Byte) || (vals[i] instanceof Short) || (vals[i] instanceof Integer)) {
		writer.writeInt(TYPE_INT);
		writer.writeInt(((Number)vals[i]).intValue());
	    }
	    else if ((vals[i] instanceof Boolean)) {
		writer.writeInt(TYPE_BOOL);
		writer.writeBoolean(((Boolean)vals[i]).booleanValue());
	    }
	    else if ((vals[i] instanceof String)) {
		String s = (String)vals[i];
		writer.writeInt(TYPE_STRING);
		writer.writeInt(s.length());
		if (s.length() > 0) writer.writeBytes(s);
	    }
	    else if ((vals[i] instanceof Float) || (vals[i] instanceof Double)) {
		long l = Double.doubleToLongBits(((Number)vals[i]).doubleValue());
		writer.writeInt(TYPE_FLOAT);
		int m = (((int)(l >> 22)) & 0x3fffffff) | 0x40000000;
		writer.writeInt(l<0 ? -m : m);
		writer.writeInt( ((int) ((l >> 52) & 0x7ff))-1023 );
	    }
	    else if (vals[i] instanceof HomeGearArray) {
		HomeGearArray a = (HomeGearArray)vals[i];
		writer.writeInt(a.isStruct() ? TYPE_STRUCT : TYPE_ARRAY);
		writer.write(a.data, a.offs, a.length);
		child[i] = a;
	    }
	    else throw new IllegalArgumentException("Unsupported type of element "+i);
	}
	data = buf.toByteArray();
	length = data.length;
	compress(data,offs);
    }

/**
  * Constructs an array from a list of variables.
  * @param vals Values of the array.
  * @throws IllegalArgumentException if a parameter is invalid.
  * @throws IOException if an I/O error occurs.
  */
    public HomeGearArray(Object... vals) throws IllegalArgumentException,IOException  {
	this(null, vals);
    }

/**
  * Constructs an array or struct from a raw data. The data is not copied.
  * @param struct True if it is a struct.
  * @param data_ Buffer containing data.
  * @param offs_ Offset of data.
  * @throws IllegalArgumentException if a parameter is invalid.
  */
  public HomeGearArray(boolean struct, byte[] data_, int offs_) throws IllegalArgumentException  {
	isNamed = struct;
	data = data_;
	offs = offs_;
	length = data.length - offs; 
	size = intEncode(offs);
	if ( size<0 ) throw new IllegalArgumentException("Invalid size");
	if ( size==0 ) {
	    length = 4;
	    valPos = null;
	    namePos = null;
	    child = null;
	    return;
	}

	valPos = new int[size];
	namePos = new int[size];
	child = new HomeGearArray[size];
	
	int pos = offs+4;
	for (int idx=0; idx<size; idx++) {
	    if ( isNamed ) {
		namePos[idx] = pos;
		pos += 4 + intEncode(pos);
	    } 
	    valPos[idx] = pos;
	    child[idx] = null;
	    int t = intEncode(pos);
	    if ( t == TYPE_INT ) pos+=8;
	    else if ( t == TYPE_BOOL ) pos+=5;
	    else if ( t == TYPE_FLOAT ) pos+=12;
	    else if ( t == TYPE_STRING || t == TYPE_BASE64 ) pos+=8+intEncode(pos+4);
	    else if ( t == TYPE_ARRAY || t == TYPE_STRUCT ) {
		child[idx] = new HomeGearArray( t == TYPE_STRUCT, data, pos+4);
		pos += 4 + child[idx].length;
	    }
	    else throw new IllegalArgumentException("Unknown type of element "+idx);
	}
	length = pos-offs;
    } 
    
/**
  * This makes a copy of the instance. 
  * If instances are created from raw data using the constructor the data buffer is re-used in order to save resources.
  * This method copies the part of data buffer which is used by the instance and creates a new one from it.
  * If only a small fraction of a large data set need to be saved this method should be used.
  * @return a copy of this instance.
  */
    public HomeGearArray copy() {
	byte[] cbuf = Arrays.copyOfRange(data, offs, offs+length);
	try {
	    return new HomeGearArray(isNamed, cbuf, 0);
	}
	catch ( Exception e ) {
	    System.err.println("HomeGearArry.copy(): Internal error");
	}
	return this;
    }

/**
  * Replaces data buffer and releases duplicate buffers of children.
  * @param data_ New data buffer.
  * @param offs_ New offset.
  */
    protected void compress(byte[] data_, int offs_) {
	int shift = offs_-offs;
	data = data_;
	offs = offs_;
	for (int i=0; i<size; i++) {
	    if ( namePos[i] > 3 ) namePos[i] += shift;
	    valPos[i] += shift;
	    if ( child[i] != null ) child[i].compress(data,valPos[i]+4);
	}
    }

/**
  * Returns number of elements.
  * @return number of elements.
  */
    public int size() {
	return size;
    }

/**
  * Returns true if it is a struct, i.e. if elements have names.
  * @return true if it is a struct.
  */
    public boolean isStruct() {
	return isNamed;
    }

/**
  * Encodes integer from data at a given position.
  * @param pos The position.
  * @return the integer.
  * @throws IllegalArgumentException if Index is out of bounds.
  */
    protected int intEncode(int pos) throws IllegalArgumentException {
	if (pos<0 || pos+4>data.length) throw new IllegalArgumentException("Unexpected and of buffer");
	return (data[pos]<<24) | ((data[pos+1] & 255)<<16) | ((data[pos+2] & 255)<<8) | (data[pos+3] & 255);
    }
    
/**
  * Returns the index of a element identified by field name.
  * Result is -1 if the HomeGearArray is not a struct or if element was not found 
  * @param name The name of the field.
  * @return the index of a element identified by field name.
  * @throws UnsupportedEncodingException if your computer does not support ASCII.
  */
    public int find(String name) throws UnsupportedEncodingException {
	if ( ! isNamed ) return -1;
	try {
	    for ( int i=0; i<size; i++ )
		if ( name.equals(name(i)) ) return i;
	}
	catch ( IndexOutOfBoundsException e ) {
	}
	return -1;
    }

/**
  * Returns the type of an element.
  * @param idx The element index.
  * @return type of the element.
  * @throws IndexOutOfBoundsException if Index is out of bounds.
  */
    public int type(int idx) throws IndexOutOfBoundsException {
	if ( (idx<0) | (idx>=size)) throw new IndexOutOfBoundsException("Index out of bounds");
	return intEncode(valPos[idx]);
    }

/**
  * Returns a boolean element.
  * @param idx The element index.
  * @throws IndexOutOfBoundsException if Index is out of bounds.
  * @throws IllegalArgumentException if element type is invalid.
  */
    public boolean boolValue(int idx) throws IndexOutOfBoundsException, IllegalArgumentException {
	if (type(idx) != TYPE_BOOL) throw new IllegalArgumentException("Element "+idx+" is not a boolean");
	return data[valPos[idx]+4] != 0;
    }

/**
  * Returns a integer element.
  * @param idx The element index.
  * @throws IndexOutOfBoundsException if Index is out of bounds.
  * @throws IllegalArgumentException if element type is invalid.
  */
    public int intValue(int idx) throws IndexOutOfBoundsException, IllegalArgumentException {
	if (type(idx) != TYPE_INT) throw new IllegalArgumentException("Element "+idx+" is not an integer");
	return intEncode(valPos[idx]+4);
    }

/**
  * Returns value of a float or integer object.
  * @param idx The element index.
  * @throws IndexOutOfBoundsException if Index is out of bounds.
  * @throws IllegalArgumentException if element type is invalid.
  */
    public double doubleValue(int idx) throws IndexOutOfBoundsException, IllegalArgumentException {
	if (type(idx) == TYPE_INT) return intValue(idx);
	if (type(idx) != TYPE_FLOAT) throw new IllegalArgumentException("Element "+idx+" is neither a float nor an integer");
	int p = valPos[idx]+4;
	return intEncode(valPos[idx]+4) * Math.pow(2,intEncode(valPos[idx]+8)-30);
    }

/**
  * Returns child array or struct.
  * @param idx the element index.
  * @throws IndexOutOfBoundsException if Index is out of bounds.
  * @throws IllegalArgumentException if element type is invalid.
  */
    public HomeGearArray child(int idx) throws IndexOutOfBoundsException, IllegalArgumentException {
	if ( type(idx)!=TYPE_ARRAY && type(idx)!=TYPE_STRUCT) throw new IllegalArgumentException("Element "+idx+" is neither an array nor a struct");
	return child[idx];
    }

/**
  * Returns a string representation of a element.
  * @param idx the element index.
  * @param lb line break string (including indent).
  * @throws IndexOutOfBoundsException if Index is out of bounds.
  * @throws IllegalArgumentException if element type is invalid.
  * @throws UnsupportedEncodingException if your computer does not support ASCII.
  */
    public String stringValue(String lb, int idx)  throws IndexOutOfBoundsException, IllegalArgumentException, UnsupportedEncodingException {
	int t = type(idx);
	if ( t == TYPE_INT ) return String.valueOf(intValue(idx));
	if ( t == TYPE_BOOL ) return String.valueOf(boolValue(idx));
	if ( t == TYPE_FLOAT ) return String.valueOf(doubleValue(idx));
	if ( t == TYPE_STRING || t == TYPE_BASE64 ) return new String(data, valPos[idx]+8, intEncode(valPos[idx]+4), "ISO-8859-1" );
	if ( t == TYPE_ARRAY || t == TYPE_STRUCT ) return child[idx].toString(lb);
	throw new IllegalArgumentException("Unknown type of element "+idx);
    }

/**
  * Returns the name of an element or null if its an array.
  * @param idx the element index.
  * @throws IndexOutOfBoundsException if Index is out of bounds.
  * @throws UnsupportedEncodingException if your computer does not support ASCII.
  */
    public String name(int idx) throws IndexOutOfBoundsException, UnsupportedEncodingException {
	if ( (idx<0) | (idx>=size)) throw new IndexOutOfBoundsException("Index out of bounds");
	int p = namePos[idx];
	if ( p<4 ) return null;
	return new String(data, namePos[idx]+4, intEncode(namePos[idx]), "ISO-8859-1" );
    }
    
/**
  * Returns a string representation.
  * @param lb line break string (including indent).
  * @return a string representation.
  */
     public String toString(String lb) {
        StringBuilder sb = new StringBuilder();
        String lb2 = new String(lb+"    ");
        sb.append("{" + (size>1 ? lb2 : "   ") );
        for (int i=0; i<size; i++) {
    	    if (i>0 ) sb.append("," + lb2);
    	    try {
    		String name = name(i);
    		if ( name !=null ) sb.append(name + " = ");
    		sb.append(stringValue(lb2,i));
    	    }
    	    catch ( Exception e ) {
    		sb.append("(unknown: "+e.getLocalizedMessage() +")" );
    	    }
    	}
        sb.append( (size > 1 ? lb : "   ") +"}");
        return sb.toString();
     }

/**
  * Returns a string representation.
  * @return a string representation.
  * @throws IllegalArgumentException if data is corrupt.
  */
     public String toString() {
    	return toString("\n");
     }


// ******* main ****************************************************************
/* * Main method which demonstrates usage of HomeGearArray.
   * @param args command line parameters
   */
    public static void main (String args[]) {
	//do some type tests
	try {
	// construct from variables
	    HomeGearArray a1 = new HomeGearArray(
		new String[] {
		    "void", 
		    "boolean", 
	    	    "integer", 
		    "float",
		    "string",
		    "array",
		    "struct"
		}, 
		new Object[] {
		    null, 
		    true, 
		    -123456789, 
		    -1.23456789e-20, 
		    "Hello world!",
		    new HomeGearArray(-987654321, 987654321, 4.567890123e100, -7.890123456e-100),
		    new HomeGearArray( new String[] {
		    	    "Field 1",
			    "Field 2"
			}, new Object[] {
			    1,
			    "B"
			}
		    )
		} 
	    );
	    System.out.println(a1);  
	    // construct from raw data
	    System.out.println(new HomeGearArray(true, a1.data, a1.offs));
	    System.out.println(new HomeGearArray(1, 2, 3, new HomeGearArray( 41 ), new HomeGearArray( new HomeGearArray( new HomeGearArray( 511, 512, 513 ) ) ) ) );
	}
	catch (Exception e) {
	    System.err.println("Error: "+e.getLocalizedMessage() );
	} 
    }
}
