package homegear;

import java.io.*;
import java.util.*;
import java.text.*;
import java.security.*;
import java.security.cert.*;
import java.net.*;
import javax.net.*;
import javax.net.ssl.*;

/**
  * This class is responsible for the communication with the homegear server.
  * It supports  
  * <ul>    
  * <li> binary RPC requests with header (ssl with authorization as recommended)
  * <li> binary responses
  * <li> error responses 
  * </ul>
  * A Usage demonstration is given in {@link #main(String[])}.
  */

public class HomeGearSocket {
    private String host, user, pw;
    private int port;
    private Socket socket;
    private OutputStream output;
    private InputStream input;

// ******* encodeBase64 *********************************************************
    private static String encodeBase64(String s) {
        return encodeBase64(s.getBytes());
    }

    private static String encodeBase64(byte[] src) {
	return encodeBase64(src, 0, src.length);
    }

    private static String encodeBase64(byte[] src, int start, int length) {
        final String charSet = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
	byte[] encodeData = new byte[64];
        byte[] dst = new byte[(length+2)/3 * 4 + length/72];
        int x = 0;
        int dstIndex = 0;
        int state = 0;
        int old = 0;
        int len = 0;
	int max = length + start;

	for (int i = 0; i<64; i++) {
	    byte c = (byte) charSet.charAt(i);
	    encodeData[i] = c;
	}
	
        for (int srcIndex = start; srcIndex<max; srcIndex++) {
	    x = src[srcIndex];
	    switch (++state) {
	    case 1:
	        dst[dstIndex++] = encodeData[(x>>2) & 0x3f];
		break;
	    case 2:
	        dst[dstIndex++] = encodeData[((old<<4)&0x30) 
	            | ((x>>4)&0xf)];
		break;
	    case 3:
	        dst[dstIndex++] = encodeData[((old<<2)&0x3C) 
	            | ((x>>6)&0x3)];
		dst[dstIndex++] = encodeData[x&0x3F];
		state = 0;
		break;
	    }
	    old = x;
	    if (++len >= 72) {
	    	dst[dstIndex++] = (byte) '\n';
	    	len = 0;
	    }
	}

	switch (state) {
	    case 1: dst[dstIndex++] = encodeData[(old<<4) & 0x30];
		dst[dstIndex++] = (byte) '=';
		dst[dstIndex++] = (byte) '=';
		break;
	    case 2: dst[dstIndex++] = encodeData[(old<<2) & 0x3c];
		dst[dstIndex++] = (byte) '=';
		break;
	}
	return new String(dst);
    }
    
// ******* constructor *********************************************************
/* * Constructs the class and establishes the connection.
   * If a certification file is given it is loaded to the keystore.
   * @param host the host name.
   * @param port the port number.
   * @param user_ the user name.
   * @param pw_ the password.
   * @param cert the SSL certification file, e.g. homegear.crt.
   * @throws IOException if an I/O error or ssl validation problem occurs
   */
    public HomeGearSocket(String host, int port, String user_, String pw_, String cert) throws IOException {
	user = user_;
	pw = pw_;

	if ( cert!=null && cert.length()>0) {
	    try {
		KeyStore ks = KeyStore.getInstance(KeyStore.getDefaultType());
		ks.load(null,null);
		ks.setCertificateEntry("homegear", CertificateFactory.getInstance("X.509").generateCertificate(new FileInputStream(cert)));

		TrustManagerFactory tmf = TrustManagerFactory.getInstance("X.509");
		tmf.init(ks);
    
		SSLContext sc = SSLContext.getInstance("SSL");
		sc.init(null, tmf.getTrustManagers(), null);
		socket = (SSLSocket) sc.getSocketFactory().createSocket(host,port);
    	    }
	    catch (Exception e) {
		throw new IOException(e.getLocalizedMessage() );
	    } 
	} else {
	    socket = SocketFactory.getDefault().createSocket(host,port);
	}

        output = socket.getOutputStream();
        input = socket.getInputStream();
    }
    
// ******* binaryRequest *******************************************************
/* with header
 * @throws IllegalArgumentException if a parmeter type is not supported.
 * @throws IOException if an I/O error occurs
 */
  
    public void request(String method, Object... params) throws IllegalArgumentException, IOException {
	ByteArrayOutputStream buf = new ByteArrayOutputStream();
	DataOutputStream header = new DataOutputStream(buf);
	
	HomeGearArray content = new HomeGearArray(params);
	
	// signature + header + content length
	header.writeBytes("Bin");				// Signature
	header.write(0x40);	
	String s = "Basic " + encodeBase64(user + ":" + pw);
//	System.out.println("Authorization String: " + s);
	header.writeInt(4+4+13+4+s.length());			// header start: length of header 
	header.writeInt(1);					// header field count
	header.writeInt(13);					// length of header field name
	header.writeBytes("Authorization");			// field name
	header.writeInt(s.length());				// value length
	header.writeBytes(s);					// value length : header end
	header.writeInt(4+method.length()+content.length);	// start of of content: content length
	header.writeInt(method.length());
	header.writeBytes(method);				// value length : header end
	
	output.write(buf.toByteArray());
	output.write(content.data, content.offs, content.length);
    } 


// ******* binaryResponse ******************************************************
/*
 * @throws IllegalArgumentException response is invalid
 * @throws IOException if an I/O error occurs
 */
    public HomeGearResponse response() throws IOException, IllegalArgumentException {
	byte[] header = new byte[8];
	input.read(header);
	if ( (header[0]!=0x42) || (header[1]!=0x69) || (header[2]!=0x6E) || ((header[3]!=1) && (header[3]!=-1)) ) 
	    throw new IllegalArgumentException("Invalid signature or type");
	int len = (header[4]<<24) | ((header[5] & 255)<<16) | ((header[6] & 255)<<8) | (header[7] & 255);
	if (len<0) throw new IllegalArgumentException("Invalid length");
	
	byte[] data = new byte[len];
	input.read(data);
	return HomeGearResponse.createResponse( header[3]==0xff, data);
    }

// ******* main ****************************************************************
/* Main method which demonstrates usage of {@link HomeGearSocket}. 
 * It can print some info, query devices and also log changes of their parameters, see parameter '-h'.
 * @param args comand line parameters
 */
    public static void main (String args[]) {
	String host = "127.0.0.1";
	int port = 2001;
	String user = "homegear";
	String password = "";
	String certFile = null;
	boolean iv = false, id = false, im = false, log=false;

	final String helpMsg = new String (
		"Parameters:\n"+
		"    -hn <address>    Host name (default: 127.0.0.1)\n" +
		"    -hp <port>       Port number (default:  2001)\n" +
		"    -u <user name>   User name (default: homegear)\n" +
		"    -p <password>    Password\n" + 
		"    -c <certificate> Certificate file for SLL connections (default: no SSL)\n " +
		"    -iv              Print homegear version\n" + 
		"    -im              Print messages\n" +
		"    -id              Print device list\n" + 
		"    -l               Log value changes\n" + 
		"    -h               Help\n" + 
		"Ordered parameters (may be specified multiple times):\n"+
		"    -ii <device id>  Print all device values" 
		);

//process the parameters
	try {
	    for (int i=0; i<args.length; i++ ) {
		if ( args[i].equals("-hn") ) {
		    i++;
		    if (i>=args.length) throw new Exception("Argument expected after -hn");
		    host = args[i];
		}
		else if ( args[i].equals("-hp") ) {
		    i++;
		    if (i>=args.length) throw new Exception("Number expected after -hp");
		    try { 
    			port = Integer.parseInt( args[i] );
		    } 
		    catch (Exception e) {
		        throw new Exception("Number expected after -hp");
		    }
		}
		else if ( args[i].equals("-u") ) {
		    i++;
		    if (i>=args.length) throw new Exception("Argument expected after -u");
		    user = args[i];
		}
		else if ( args[i].equals("-p") ) {
		    i++;
		    if (i>=args.length) throw new Exception("Argument expected after -p");
		    password = args[i];
		}
		else if ( args[i].equals("-c") ) {
		    i++;
		    if (i>=args.length) throw new Exception("Argument expected after -c");
		    certFile = args[i];
		}
		else if ( args[i].equals("-iv") ) {
		    iv = true;
		}
		else if ( args[i].equals("-id") ) {
		    id = true;
		}
		else if ( args[i].equals("-im") ) {
		    im = true;
		}
		else if ( args[i].equals("-ii") ) {
		    i++;
		}
		else if ( args[i].equals("-l") ) {
		    log = true;
		}
		else if ( args[i].equals("-h") ) {
		    System.err.println(helpMsg);
		    System.exit(0);
		}
		else throw new Exception ("Invalid parameter: "  + args[i] );
	    }
	}
	catch ( Exception e ) {
	    System.err.println("Error: "+e.getLocalizedMessage() );
	    System.err.println(helpMsg);
	    System.exit(1);
	}
	    
	// connect to server and run some other tests
	try {
	    HomeGearSocket hg = new HomeGearSocket(host, port, user, password, certFile);
	    if ( iv ) {
		hg.request("getVersion");
		System.out.println("Version = " + hg.response());
	    }
	    if ( im ) {
		hg.request("getServiceMessages", true);
		System.out.println("ServiceMessages = " + hg.response());
	    }
	    if ( id ) {
		hg.request("listDevices", false, new HomeGearArray( "ID", "ADDRESS", "TYPE") );
		HomeGearArray devices = hg.response();
		System.out.println("Devices = {");
		for ( int i=0; i<devices.size(); i++) {
		    if ( devices.type(i) == HomeGearArray.TYPE_STRUCT ) {
			try {
			    HomeGearArray info = devices.child(i);
			    int j = info.find("ID");
			    int k = info.intValue(j);
			    hg.request("getDeviceInfo",k, new HomeGearArray( "NAME", "INTERFACE" ) );
			    HomeGearArray name = hg.response();
			    if ( j >=0  ) {
				System.out.println("    {");
				System.out.print("            ID = " + k);
				int l = name.find( "NAME" );
				if ( l>=0 ) System.out.print(",\n            NAME = " + name.stringValue("\n            ",l));
				
				for (l=0; l<info.size(); l++ ) {
				    if ( !info.name(l).equals("ID") ) {
					System.out.print(",\n            "+info.name(l) + " = " + info.stringValue("\n            ",l));
				    }
				}

				l = name.find( "INTERFACE" );
				if ( l>=0 ) System.out.print(",\n            INTERFACE = " + name.stringValue("\n            ",l));
				System.out.println("\n    }" + ( i<devices.size()-1 ? "," : "" ) );
			    }
			}
			catch ( Exception e ) {
			    System.out.println("Warning: cannot parse device info: " + e.getLocalizedMessage() );
			}
		    }
		}
		System.out.println("}");
	    }
	    
	    // ordered prams
	    for (int i=0; i<args.length; i++ ) {
		if ( args[i].equals("-ii") ) {
		    int ID;
		    if (i>=args.length) throw new Exception("Device ID expected after "+args[i]);
		    try { 
    			ID = Integer.parseInt( args[i+1] );
		    } 
		    catch (Exception e) {
		        throw new Exception("Device ID expected after "+args[i]);
		    }
		    if ( args[i].equals("-ii") ) {
			hg.request("getAllValues", ID);
			System.out.println("values = " + hg.response());
			
		    }
		    i++;
		}		
	    }
	    
	    if ( log ) {
		SimpleDateFormat logDateFormat = new SimpleDateFormat("yyyyMMdd'T'HH:mm:ss");
		Vector<HomeGearName> peerNames = new Vector<HomeGearName>();
		Vector<HomeGearParameter> params = new Vector<HomeGearParameter>();
		while ( true ) {
		    hg.request("getAllValues");
		    HomeGearArray devices = hg.response();
		
		    for ( int i=0; i<devices.size(); i++) {
			if ( devices.type(i) == HomeGearArray.TYPE_STRUCT ) {
			    try {
				HomeGearArray dev = devices.child(i);
				String peerName = null;
				int j = dev.find("ID");
				int ID = dev.intValue(j);
				// get peer name
				j = peerNames.indexOf(new HomeGearName(ID,null));
				if ( j < 0 ) {
				    try {
					hg.request("getDeviceInfo",ID, new HomeGearArray( "NAME" ) );
					HomeGearArray name = hg.response();
					peerName = name.stringValue("",name.find( "NAME" ));
				    } catch (Exception e) {
					peerName = "";
				    }
				    peerNames.add(new HomeGearName(ID,peerName));
//				    System.out.println("got peer name for ID " + ID + ": " + peerName );
				} else {
				    peerName = peerNames.elementAt(j).name;
				}
				// read channels
			    	HomeGearArray channels = dev.child(dev.find("CHANNELS"));
				for ( j = 0; j<channels.size(); j++ ) {
				    HomeGearArray chan = channels.child(j);
				    int CI = chan.intValue(chan.find("INDEX"));
				    HomeGearArray paramSet = chan.child(chan.find("PARAMSET"));
				    for ( int k = 0; k<paramSet.size(); k++ ) {
					int l = params.indexOf( new HomeGearParameter(ID, CI, paramSet.name(k), peerName,null) );
					if ( l < 0 ) {
					    HomeGearParameter e = new HomeGearParameter(ID, CI, paramSet.name(k), peerName, paramSet.child(k));
					    params.add(e);
					    System.out.println("(new) " +logDateFormat.format(new Date()) + ": " + e);
					}
					else {
					    HomeGearParameter e = params.elementAt(l);
					    if ( e.update(paramSet.child(k)) ) System.out.println(logDateFormat.format(new Date()) + ": " + e);
					}
				    } 
				}  
			    } 
			    catch ( Exception e) {
				System.err.println("Error: "+e.getLocalizedMessage() );
			    } 
			}
		    }
		    try { Thread.sleep( 1000 ); } catch (Exception e) { }
		}
	    }
	}
	catch (Exception e) {
	    System.err.println("Error: "+e.getLocalizedMessage() );
	} 
    }
    	
   
}
